# -*- coding: utf-8 -*-
#
# Copyright (C) 2018, 2019, 2020 Esteban J. G. Gabancho.
#
# Invenio-S3 is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""S3 file storage support for Invenio."""

S3_ENDPOINT_URL = None
"""S3 server URL endpoint.

If using Amazon AWS S3 service this config variable can be set to None as the
underlining library, `boto3 <https://boto3.readthedocs.io/en/latest/>`_,
will automatically construct the appropriate URL to use when communicating with
a service.

If set to a value (including the "http/https" scheme) it will be passed as
``endpoint_url`` to boto3 `client
<https://boto3.readthedocs.io/en/latest/reference/core/session.html#boto3.session.Session.client>`_.
"""

S3_REGION_NAME = None
"""S3 region name

This is entirely optional, and if not provided, the region name will be
automatically set to 'us-east-1'.

If set to a value it will be passed as ``region_name`` to boto3 `client
<https://boto3.readthedocs.io/en/latest/reference/core/session.html#boto3.session.Session.client>`_.
"""

S3_ACCESS_KEY_ID = None
"""The access key to use when creating the client.

This is entirely optional, and if not provided, the credentials configured for
the session will automatically be used.
See `Configuring Credentials
<https://boto3.readthedocs.io/en/latest/guide/configuration.html#credentials>`_.
for more information.
"""

S3_SECRET_ACCESS_KEY = None
"""The secret key to use when creating the client.

This is entirely optional, and if not provided, the credentials configured for
the session will automatically be used.
See `Configuring Credentials
<https://boto3.readthedocs.io/en/latest/guide/configuration.html#credentials>`_.
for more information.
"""

S3_URL_EXPIRATION = 60
"""Number of seconds the file serving URL will be valid.

See `Amazon Boto3 documentation on presigned URLs
<https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.generate_presigned_url>`_
for more information.
"""

S3_SIGNATURE_VERSION = 's3v4'
"""Version of the S3 signature algorithm. Can be 's3' (v2) or 's3v4' (v4).
See `Amazon Boto3 documentation on configuration variables
<https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#configuration-file>`_
for more information.
"""

S3_MAXIMUM_NUMBER_OF_PARTS = 10000
"""Maximum number of parts to be used.
See `AWS Multipart Upload Overview
<https://docs.aws.amazon.com/AmazonS3/latest/dev/mpuoverview.html>`_ for more
information.
"""

S3_DEFAULT_BLOCK_SIZE = 5 * 2**20
"""Default block size value used to send multi-part uploads to S3.
Typically 5Mb is minimum allowed by the API."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018, 2019 Esteban J. G. Gabancho.
#
# Invenio-S3 is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""S3 file storage support for Invenio."""

from __future__ import absolute_import, print_function

import warnings

import boto3
from flask import current_app
from werkzeug.utils import cached_property

from . import config


class InvenioS3(object):
    """Invenio-S3 extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    @cached_property
    def init_s3fs_info(self):
        """Gather all the information needed to start the S3FSFileSystem."""
        if 'S3_ACCCESS_KEY_ID' in current_app.config:
            current_app.config['S3_ACCESS_KEY_ID'] = current_app.config[
                'S3_ACCCESS_KEY_ID']
            warnings.warn(
                'Key S3_ACCCESS_KEY_ID contained a typo and has been '
                'corrected to S3_ACCESS_KEY_ID, support for the '
                'flawed version will be removed.',
                DeprecationWarning
            )

        if 'S3_SECRECT_ACCESS_KEY' in current_app.config:
            current_app.config['S3_SECRET_ACCESS_KEY'] = current_app.config[
                'S3_SECRECT_ACCESS_KEY']
            warnings.warn(
                'Key S3_SECRECT_ACCESS_KEY contained a typo and has been '
                'corrected to S3_SECRET_ACCESS_KEY, support for the '
                'flawed version will be removed.',
                DeprecationWarning
            )

        info = dict(
            key=current_app.config.get('S3_ACCESS_KEY_ID', ''),
            secret=current_app.config.get('S3_SECRET_ACCESS_KEY', ''),
            client_kwargs={},
            config_kwargs={
                's3': {
                    'addressing_style': 'path',
                },
                'signature_version': current_app.config.get(
                    'S3_SIGNATURE_VERSION', 's3v4'
                ),
            },
        )

        s3_endpoint = current_app.config.get('S3_ENDPOINT_URL', None)
        if s3_endpoint:
            info['client_kwargs']['endpoint_url'] = s3_endpoint

        region_name = current_app.config.get('S3_REGION_NAME', None)
        if region_name:
            info['client_kwargs']['region_name'] = region_name

        return info

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        app.extensions['invenio-s3'] = self

    def init_config(self, app):
        """Initialize configuration."""
        for k in dir(config):
            if k.startswith('S3_'):
                app.config.setdefault(k, getattr(config, k))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018, 2019 Esteban J. G. Gabancho.
#
# Invenio-S3 is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Module tests."""

from __future__ import absolute_import, print_function

from invenio_s3 import InvenioS3


def test_version():
    """Test version import."""
    from invenio_s3 import __version__
    assert __version__


def test_init(appctx):
    """Test extension initialization."""
    assert 'invenio-s3' in appctx.extensions

    appctx.config['S3_ENDPOINT_URL'] = 'https://example.com:1234'
    appctx.config['S3_REGION_NAME'] = 'eu-west-1'
    s3_connection_info = appctx.extensions['invenio-s3'].init_s3fs_info
    assert s3_connection_info['client_kwargs'][
        'endpoint_url'] == 'https://example.com:1234'
    assert s3_connection_info['client_kwargs'][
        'region_name'] == 'eu-west-1'


def test_access_key(appctx):
    """Test correct access key works together with flawed one."""
    appctx.config['S3_ACCCESS_KEY_ID'] = 'secret'
    try:
        # Delete the cached value in case it's there already
        del appctx.extensions['invenio-s3'].__dict__['init_s3fs_info']
    except KeyError:
        pass
    s3_connection_info = appctx.extensions['invenio-s3'].init_s3fs_info
    assert s3_connection_info['key'] == 'secret'


def test_secret_key(appctx):
    """Test correct secret key works together with flawed one."""
    appctx.config['S3_SECRECT_ACCESS_KEY'] = 'secret'
    try:
        # Delete the cached value in case it's there already
        del appctx.extensions['invenio-s3'].__dict__['init_s3fs_info']
    except KeyError:
        pass
    s3_connection_info = appctx.extensions['invenio-s3'].init_s3fs_info
    assert s3_connection_info['secret'] == 'secret'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018, 2019, 2020, 2021 Esteban J. G. Gabancho.
#
# Invenio-S3 is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""S3 file storage interface."""

from functools import partial, wraps
from math import ceil

import s3fs
from flask import current_app
from invenio_files_rest.errors import StorageError
from invenio_files_rest.storage import PyFSFileStorage, pyfs_storage_factory

from .helpers import redirect_stream


def set_blocksize(f):
    """Decorator to set the correct block size according to file size."""
    @wraps(f)
    def inner(self, *args, **kwargs):
        size = kwargs.get('size', None)
        block_size = (
            ceil(size / current_app.config['S3_MAXIMUM_NUMBER_OF_PARTS'])
            if size
            else current_app.config['S3_DEFAULT_BLOCK_SIZE']
        )

        if block_size > self.block_size:
            self.block_size = block_size
        return f(self, *args, **kwargs)

    return inner


class S3FSFileStorage(PyFSFileStorage):
    """File system storage using Amazon S3 API for accessing files."""

    def __init__(self, fileurl, **kwargs):
        """Storage initialization."""
        self.block_size = current_app.config['S3_DEFAULT_BLOCK_SIZE']
        super(S3FSFileStorage, self).__init__(fileurl, **kwargs)

    def _get_fs(self, *args, **kwargs):
        """Get PyFilesystem instance and S3 real path."""
        if not self.fileurl.startswith('s3://'):
            return super(S3FSFileStorage, self)._get_fs(*args, **kwargs)

        info = current_app.extensions['invenio-s3'].init_s3fs_info
        fs = s3fs.S3FileSystem(default_block_size=self.block_size, **info)

        return (fs, self.fileurl)

    @set_blocksize
    def initialize(self, size=0):
        """Initialize file on storage and truncate to given size."""
        fs, path = self._get_fs()

        if fs.exists(path):
            fp = fs.rm(path)
        fp = fs.open(path, mode='wb')

        try:
            to_write = size
            fs_chunk_size = fp.blocksize  # Force write every time
            while to_write > 0:
                current_chunk_size = (
                    to_write if to_write <= fs_chunk_size else fs_chunk_size
                )
                fp.write(b'\0' * current_chunk_size)
                to_write -= current_chunk_size
        except Exception:
            fp.close()
            self.delete()
            raise
        finally:
            fp.close()

        self._size = size

        return self.fileurl, size, None

    def delete(self):
        """Delete a file."""
        fs, path = self._get_fs()
        if fs.exists(path):
            fs.rm(path)
        return True

    @set_blocksize
    def update(
        self,
        incoming_stream,
        seek=0,
        size=None,
        chunk_size=None,
        progress_callback=None,
    ):
        """Update a file in the file system."""
        old_fp = self.open(mode='rb')
        updated_fp = S3FSFileStorage(self.fileurl, size=self._size).open(
            mode='wb'
        )
        try:
            if seek >= 0:
                to_write = seek
                fs_chunk_size = updated_fp.blocksize
                while to_write > 0:
                    current_chunk_size = (
                        to_write
                        if to_write <= fs_chunk_size
                        else fs_chunk_size
                    )
                    updated_fp.write(old_fp.read(current_chunk_size))
                    to_write -= current_chunk_size

            bytes_written, checksum = self._write_stream(
                incoming_stream,
                updated_fp,
                chunk_size=chunk_size,
                size=size,
                progress_callback=progress_callback,
            )

            if (bytes_written + seek) < self._size:
                old_fp.seek((bytes_written + seek))
                to_write = self._size - (bytes_written + seek)
                fs_chunk_size = updated_fp.blocksize
                while to_write > 0:
                    current_chunk_size = (
                        to_write
                        if to_write <= fs_chunk_size
                        else fs_chunk_size
                    )
                    updated_fp.write(old_fp.read(current_chunk_size))
                    to_write -= current_chunk_size
        finally:
            old_fp.close()
            updated_fp.close()

        return bytes_written, checksum

    def send_file(
        self,
        filename,
        mimetype=None,
        restricted=True,
        checksum=None,
        trusted=False,
        chunk_size=None,
        as_attachment=False,
    ):
        """Send the file to the client."""
        try:
            fs, path = self._get_fs()
            s3_url_builder = partial(
                fs.url, path, expires=current_app.config['S3_URL_EXPIRATION']
            )

            return redirect_stream(
                s3_url_builder,
                filename,
                mimetype=mimetype,
                restricted=restricted,
                trusted=trusted,
                as_attachment=as_attachment,
            )
        except Exception as e:
            raise StorageError('Could not send file: {}'.format(e))

    @set_blocksize
    def copy(self, src, *args, **kwargs):
        """Copy data from another file instance.

        If the source is an S3 stored object the copy process happens on the S3
        server side, otherwise we use the normal ``FileStorage`` copy method.
        """
        if src.fileurl.startswith('s3://'):
            fs, path = self._get_fs()
            fs.copy(src.fileurl, path)
        else:
            super(S3FSFileStorage, self).copy(src, *args, **kwargs)

    @set_blocksize
    def save(self, *args, **kwargs):
        """Save incoming stream to storage.

        Just overwrite parent method to allow set the correct block size.
        """
        return super(S3FSFileStorage, self).save(*args, **kwargs)


def s3fs_storage_factory(**kwargs):
    """File storage factory for S3."""
    return pyfs_storage_factory(filestorage_class=S3FSFileStorage, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018, 2019, 2020 Esteban J. G. Gabancho.
#
# Invenio-S3 is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""S3 file storage support for Invenio.

To use this module together with Invenio-Files-Rest there are a few things you
need to keep in mind.

The storage factory configuration variable, ``FILES_REST_STORAGE_FACTORY``
needs to be set to ``'invenio_s3.s3fs_storage_factory'`` importable string.

We think the best way to use this module is to have one `Localtion
<https://invenio-files-rest.readthedocs.io/en/latest/api.html#module-invenio_files_rest.models>`_
for each S3 bucket. This is just for simplicity, it can used however needed.

When creating a new location which will use the S3 API, the URI needs to start
with ``s3://``, for example
``invenio files location s3_default s3://my-bucket --default`` will
create a new location, set it as default location for your instance and use the
bucket ``my-bucket``. For more information about this command check
`Invenio-Files-Rest <https://invenio-files-rest.readthedocs.io/en/latest/>`_
documentation.

Then, there are a few configuration variables that need to be set on your
instance, like the endpoint, the access key and the secret access key, see a
more detailed description in :any:`configuration`.

.. note::

  This module doesn't create S3 buckets automatically, so before starting they
  need to be created.

  You might also want to set the correct `CORS configuration
  <https://docs.aws.amazon.com/AmazonS3/latest/dev/cors.html>`_  so files can
  be used by your interface for things like previewing a PDF with some
  Javascript library.

"""

from __future__ import absolute_import, print_function

from .ext import InvenioS3
from .storage import S3FSFileStorage, s3fs_storage_factory
from .version import __version__

__all__ = (
    '__version__',
    'InvenioS3',
    'S3FSFileStorage',
    's3fs_storage_factory',
)
